# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: https://foglamp-foglamp-documentation.readthedocs-hosted.com
# FOGLAMP_END

""" Convert a stream of time series data into a stream of files which 
    contain pd frames of snapshots of the data

    Timeseries data may come in as asset contents (filter) or from
    file data in various forms (south).
"""


__author__ = "Douglas Orr"
__copyright__ = "Copyright (c) 2020 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

import os

from datetime import datetime

import logging
from foglamp.common import logger
_LOGGER = logger.setup(__name__, level=logging.DEBUG)

from foglamp.common.plugin_helpers import HandleMap, PluginHandle

import filter_ingest

_module_dir = os.path.dirname(__file__)
_handle_map = HandleMap(__name__)

FOGLAMP_ROOT = os.environ.get("FOGLAMP_ROOT", "/usr/local/foglamp")
FOGLAMP_DATA = os.path.join(FOGLAMP_ROOT, "data")


class DataframeHandle(PluginHandle):
    RPC_SERVER_NAME = "dataframe_server"          # name of RPC server module
    RPC_CONFIG_MEMBERS = ["sample_length", "sample_offset", "output_dir"]

    def __init__(self, config, ingest_ref, callback):
        super().__init__("data-frame", (ingest_ref, callback))

        # slots
        self.enable = None
        self.input_asset = None
        self.output_asset = None
        self.forward_inputs = None
        self.sampling_rate = None
        self.frame_duration = None
        self.frame_offset = None

        self.rpc = None
        self.times = []
        
        self.reconfigure(config)
        
    def reconfigure(self, new_config):
        """ reconfigure - make the current plugin state consistent with the new configuration """

        # our server state is pretty simple, just tell the plugin to re-form its filters
        self.config_update(new_config)
        
        if self.frame_offset > self.frame_duration: raise ValueError("frame offset must be <= frame duration")
        
        self.sample_length = int(self.sampling_rate * (self.frame_duration / 1000.0))
        if self.frame_offset == 0:
            # each frame has all new data
            self.sample_offset = self.sample_length
        else:
            # index in a (small) amount
            self.sample_offset = int(self.sampling_rate * (self.frame_offset / 1000.0))

        self.output_dir = os.path.join(FOGLAMP_DATA, self.output_asset)
        os.makedirs(self.output_dir, exist_ok=True)

        self.timestamp = None

        self.rpc_setup(new_config, _module_dir)

    def _time(self):
        self.times.append(datetime.utcnow())
        if len(self.times) > 10:
            for t in self.times:
                self.logger.debug("ingest: done! {}".format(t))
            self.times = []

    def ingest(self, data):
        """ ingest - convert time series data into data frames """

        # collect up a window's worth, save it as a pickle file, publish that file name
        # ASSUME you'll get a consistent set of data columns with equal rows of data

        # self._time()

        if type(data) is dict:
            data = [data]

        # if you are forwarding inputs, start your list with your inputs
        _results = [] if not self.forward_inputs else data

        if self.enable:
            # process...
            for d in data:
                if d["asset"] != self.input_asset:
                    self.logger.debug("skipping asset {}".format(d["asset"]))
                    continue

                if not self.timestamp:
                    # save the first timestamp you see
                    self.timestamp = d.get("user_ts", d.get("timestamp"))
                    self.rpc.set_timestamp(self.timestamp)

                _fname = self.rpc.ingest(d['readings'])
                if _fname:
                    # when you have a frame's data saved, tell upstream
                    _results.append({"asset": self.output_asset,
                                     "timestamp": self.timestamp,
                                     "readings": {
                                         "datafmt": "FRAME",
                                         "dir": self.output_dir,
                                         "name": [_fname],
                                         "sampling_rate": self.sampling_rate
                                     }})
                    self.timestamp = None
                    
        filter_ingest.filter_ingest_callback(self.callback, self.ingest_ref, _results)
            

#
#
### PLUGIN API ###
#
#

_DEFAULT_CONFIG = {
    "plugin": {
        "description": "Dataframe asset consolidation plugin",
        "type": "string",
        "default": "dataframe",
        "readonly": "true"
    },
    "enable": {
        "description": "enable/disable this filter",
        "type": "boolean",
        "default": "true",
        "displayName": "Enable/disable",
        "order": "1"
    },
    "inputAsset": {
        "description": "Input asset name",
        "type": "string",
        "default": "in-timeseries",
        "displayName": "Asset name to monitor",
        "order": "2"
    },
    "outputAsset": {
        "description": "Output asset name, where frames are published",
        "type": "string",
        "default": "out-timeseries",
        "displayName": "Output asset",
        "order": "3"
    },
    "forwardInputs": {
        "description": "Forward input data through filter",
        "type": "boolean",
        "default": "false",   # default is convert, don't forward
        "displayName": "Forward input data",
        "order": "4"
    },
    # dataframe specific entries
    "samplingRate": {
        "description": "Input sampling rate",
        "type": "integer",
        "default": "8000",
        "displayName": "Sampling rate",
        "order": "5"
    },
    "frameDuration": {
        "description": "Frame size (in milliseconds of samples)",
        "type": "integer",
        "default": "1000",
        "displayName": "MS of samples",
        "order": "6"
    },
    "frameOffset": {
        "description": "Frame offset from previous frame (in milliseconds of samples)",
        "type": "integer",
        "default": "0",
        "displayName": "Offset from previous frame",
        "order": "6"
    },
}

def plugin_info():
    """ Returns information about the plugin
    Args:
    Returns:
        dict: plugin information
    Raises:
    """
    _LOGGER.debug("plugin_info called")
    return {
        "name": "dataframe",
        "version": "1.9.0",
        "mode": "none",
        "type": "filter",
        "interface": "1.0",
        "config": _DEFAULT_CONFIG
    }

def plugin_init(config, ingest_ref, callback):
    """ Initialise the plugin
    Args:
        config: JSON configuration document for the Filter plugin configuration category
        ingest_ref:
        callback:
    Returns:
        data: JSON object to be used in future calls to the plugin
    Raises:
    """
    _handle = DataframeHandle(config, ingest_ref, callback)
    _handle.rpc.plugin_init(config)
    
    _h = _handle_map.new_handle(_handle, config)
    return _h

def plugin_ingest(h, data):
    """ Absorb run signal processing algorithms on time series data "data"
    Args:
        handle: handle returned by the plugin initialisation call
        data: readings data
    """

    _handle = _handle_map.get_handle(h)
    _handle.ingest(data)

def plugin_reconfigure(h, new_config):
    """ Reconfigures the plugin

    Args:
        h: handle returned by the plugin initialisation call
        new_config: JSON object representing the new configuration category for the category
    Returns:
        new_handle: new handle to be used in the future calls
    """
    _handle = _handle_map.get_handle(h)
    _handle.reconfigure(new_config)

    h['config'].update(new_config)
    return h

def plugin_shutdown(h):
    """ Shutdowns the plugin doing required cleanup.
    Args:
        handle: handle returned by the plugin initialisation call
    Returns:
        plugin shutdown
    """
    _LOGGER.info('plugin_shutdown called')

    _handle = _handle_map.get_handle(h)
    _handle.shutdown()         # shutdown any RPC objects
    
    _handle_map.del_handle(h)  # remove our mapping
