# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: https://foglamp-foglamp-documentation.readthedocs-hosted.com
# FOGLAMP_END

""" Convert a stream of time series data into a stream of files which 
    contain pd frames of snapshots of the data

    Timeseries data may come in as asset contents (filter) or from
    file data in various forms (south).
"""


__author__ = "Douglas Orr"
__copyright__ = "Copyright (c) 2020 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

import os

import pandas as pd

import dateutil

from foglamp.common.plugin_helpers import PluginRPCServer

import logging
from foglamp.common import logger
_LOGGER = logger.setup(__name__, level=logging.DEBUG)


# (dataframe server exists to encapsulate numpy, which, in current form, can't
# exist in multiple python sub-interpreters... can be converted to client-only
# when this is fixed [by python/numpy])

FORMAT_TS="%Y-%m-%d %H:%M:%S.%f"

class DataframeServer(PluginRPCServer):
    def __init__(self):
        super().__init__('df-svc')

        # slots
        self.timestamp = None
        self.sample_length = None
        self.output_dir = None

        self.logger = _LOGGER  # XXX temp, something's wrong with inheriting the default destination

    def plugin_init(self, config):
        super().config_update(config)
        
        self.data = []

    def set_timestamp(self, timestamp):
        self.timestamp = timestamp

    def ingest(self, reading):
        _fname = None

        self.data.append(reading)

        if len(self.data) >= self.sample_length:
            # push out data collected so far into a dataframe file
            _fname, self.data = self._flush(self.data)
            
        return _fname

    def _flush(self, data):
        """ flush - write out the current frame in a new pickle file """

        def _gen_filename(pfx, ts, *args, suff="jpg"):
            """ gen_filename - generate a new, unique file name """
            _seq = 0
            while True:
                _fname = "{}_{}_{}".format(pfx, ts.strftime(FORMAT_TS), _seq)
                for a in args: _fname += "_{}".format(a)
                _fname += "."+suff
                if not os.path.exists(os.path.join(self.output_dir, _fname)):
                    break
                _seq += 1
            return _fname

        _data = data[: self.sample_length]
        _rem = data[self.sample_offset :]

        _timestamp = dateutil.parser.parse(self.timestamp)

        _fname = _gen_filename("TS", _timestamp, len(_data), suff="tspkl")
        _path = os.path.join(self.output_dir, _fname)

        # write out a window's worth of data
        _frame = pd.DataFrame(_data)
        _frame.to_pickle(_path)

        return _fname, _rem

    def plugin_shutdown(self):
        pass

server = DataframeServer()
server.serve()
